﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using Be.Timvw.Framework.ComponentModel;
using RSN.Tools;

namespace RSN.DotP
{
	public class GameDirectory
	{
		private string m_strGameDirectory;
		private HashSet<WadBase> m_hsetWads;
		private SortableBindingList<CardInfo> m_lstCards;
		private SortableBindingList<Deck> m_lstDecks;
		private Dictionary<string, WadBase> m_dicImages;
		private Dictionary<string, TdxWrapper> m_dicCachedImages;
		private Dictionary<string, Dictionary<string, string>> m_dicStringTable;
		private Dictionary<int, string> m_dicUsedIds;
		// This is a dictionary rather than a list so we can search for personalities by name (uppercase).
		private Dictionary<string, AiPersonality> m_dicPersonalities;
		private SortableBindingList<string> m_lstMusic;

		public GameDirectory(string strDir)
		{
			m_strGameDirectory = strDir;
		}

		public string GameDir
		{
			get { return m_strGameDirectory; }
		}

		public SortableBindingList<CardInfo> Cards
		{
			get { return m_lstCards; }
		}

		public SortableBindingList<Deck> Decks
		{
			get { return m_lstDecks; }
		}

		public Dictionary<string, Dictionary<string, string>> StringTable
		{
			get { return m_dicStringTable; }
		}

		public Dictionary<int, string> UsedIds
		{
			get { return m_dicUsedIds; }
		}

		public Dictionary<string, AiPersonality> Personalities
		{
			get { return m_dicPersonalities; }
		}

		public SortableBindingList<string> Music
		{
			get { return m_lstMusic; }
		}

		public CardInfo GetCardByFileName(string strFileName)
		{
			CardInfo ciCard = null;

			foreach (CardInfo ciTest in m_lstCards)
			{
				if (ciTest.Filename.Equals(strFileName, StringComparison.OrdinalIgnoreCase))
				{
					ciCard = ciTest;
					break;
				}
			}

			return ciCard;
		}

		public Deck GetDeckById(int nUid)
		{
			Deck dkDeck = null;

			foreach (Deck dkTest in m_lstDecks)
			{
				if (dkTest.Uid == nUid)
				{
					dkDeck = dkTest;
					break;
				}
			}

			return dkDeck;
		}

		public TdxWrapper LoadImage(string strImage)
		{
			if (strImage != null)
			{
				strImage = strImage.ToUpper();

				if (m_dicCachedImages.ContainsKey(strImage))
					return m_dicCachedImages[strImage];

				if (m_dicImages.ContainsKey(strImage))
				{
					TdxWrapper tdx = m_dicImages[strImage].LoadImage(strImage);
					if (Settings.GetSetting("MaintainImageCache", Settings.MAINTAIN_IMAGE_CACHE_DEFAULT))
					{
						m_dicCachedImages.Add(strImage, tdx);
					}
					return tdx;
				}
				else
					return null;
			}
			else
				return null;
		}

		private void MergeInStringTable(Dictionary<string, Dictionary<string, string>> dicToMerge)
		{
			foreach (KeyValuePair<string, Dictionary<string, string>> kvString in dicToMerge)
			{
				if (!m_dicStringTable.ContainsKey(kvString.Key))
					m_dicStringTable.Add(kvString.Key, kvString.Value);
			}
		}

		public string GetLocalizedString(string strId)
		{
			strId = strId.Trim().ToUpper();
			if (m_dicStringTable != null)
			{
				if (m_dicStringTable.ContainsKey(strId))
				{
					Dictionary<string, string> dicText = m_dicStringTable[strId];
					if (dicText.ContainsKey(Settings.LanguageCode))
						return dicText[Settings.LanguageCode];
					else if (dicText.ContainsKey("en-US"))
						return dicText["en-US"];
					else
						return "[" + strId + "]";
				}
				else
					return "[" + strId + "]";
			}
			else
				return "[" + strId + "]";
		}

		public void LoadMusic()
		{
			string strDir = m_strGameDirectory;
			if (!strDir.EndsWith("\\"))
				strDir += "\\";

			m_lstMusic = new SortableBindingList<string>();
			if (Directory.Exists(strDir + "Music"))
			{
				IEnumerable<string> files = Directory.EnumerateFiles(strDir + "Music", "*.ogg");
				foreach (string strFile in files)
					m_lstMusic.Add(Path.GetFileNameWithoutExtension(strFile));
			}
		}

		public void LoadWads()
		{
			m_dicStringTable = new Dictionary<string, Dictionary<string, string>>();
			m_dicCachedImages = new Dictionary<string, TdxWrapper>();
			m_dicUsedIds = new Dictionary<int, string>();

			if (Directory.Exists(m_strGameDirectory))
			{
				// Since the game tends to load packed WADs before unpacked ones we will do the same.

				m_hsetWads = new HashSet<WadBase>();
				m_lstCards = new SortableBindingList<CardInfo>();
				m_dicImages = new Dictionary<string, WadBase>();

				// Check for packed WADs (actual WAD files) that the game might actually load.
				//	It seems the game loads those that start with DATA_ or DECK_ (case insensitive).
				IEnumerable<string> files = Directory.EnumerateFiles(m_strGameDirectory, "*.wad", SearchOption.TopDirectoryOnly);
				foreach (string strFile in files)
				{
					if ((Path.GetFileName(strFile).StartsWith("data", StringComparison.OrdinalIgnoreCase)) ||
						(Path.GetFileName(strFile).StartsWith("deck", StringComparison.OrdinalIgnoreCase)))
					{
						WadWrapper wad = new WadWrapper(strFile, this);
						m_hsetWads.Add(wad);
						foreach (CardInfo ciCard in wad.Cards)
							m_lstCards.Add(ciCard);
						foreach (string strImage in wad.Images)
						{
							if (!m_dicImages.ContainsKey(strImage))
								m_dicImages.Add(strImage, wad);
						}
						MergeInStringTable(wad.StringTable);
					}
				}

				// Check for unpack WADs (directories with Header XMLs).
				//	It seems the game loads those that start with DATA_ or DECK_ (case insensitive).
				//	It also seems the game has issues using the directory DATA_DECKS because there may be several "wads" there so we won't load that one.
				IEnumerable<string> dirs = Directory.EnumerateDirectories(m_strGameDirectory);
				foreach (string strDir in dirs)
				{
					if (((Path.GetFileName(strDir).StartsWith("data", StringComparison.OrdinalIgnoreCase)) ||
						(Path.GetFileName(strDir).StartsWith("deck", StringComparison.OrdinalIgnoreCase))) &&
						(!Path.GetFileName(strDir).Equals("data_decks", StringComparison.OrdinalIgnoreCase)))
					{
						if (File.Exists(strDir + @"\header.xml"))
						{
							WadDirectory wad = new WadDirectory(strDir, this);
							m_hsetWads.Add(wad);
							foreach (CardInfo ciCard in wad.Cards)
								m_lstCards.Add(ciCard);
							foreach (string strImage in wad.Images)
							{
								if (m_dicImages.ContainsKey(strImage))
									m_dicImages[strImage] = wad;
								else
									m_dicImages.Add(strImage, wad);
							}
							MergeInStringTable(wad.StringTable);
						}
					}
				}

				// We should load our personalities before loading the decks so that the decks can get associated personalities.
				m_dicPersonalities = new Dictionary<string, AiPersonality>();
				foreach (WadBase wad in m_hsetWads)
				{
					wad.LoadPersonalities(this);
					foreach (AiPersonality ap in wad.AiPersonalities)
					{
						if (!m_dicPersonalities.ContainsKey(ap.Filename.ToUpper()))
							m_dicPersonalities.Add(ap.Filename.ToUpper(), ap);
					}
				}

				// Now that we have loaded all the "Wads" now we load the decks from those wads.
				m_lstDecks = new SortableBindingList<Deck>();
				foreach (WadBase wad in m_hsetWads)
				{
					wad.LoadDecks(this);
					foreach (Deck deck in wad.Decks)
					{
						m_lstDecks.Add(deck);
						if (m_dicUsedIds.ContainsKey(deck.Uid))
							Settings.ReportError(null, ErrorPriority.Medium, deck.FileName + " uses a UID that conflicts with " + m_dicUsedIds[deck.Uid] + ".");
						else
							m_dicUsedIds.Add(deck.Uid, deck.FileName);
					}
				}

				// All decks are loaded so try and associate all of their land pools.
				foreach (Deck deck in m_lstDecks)
				{
					if (deck.IsLandPool)
					{
						// This is a land pool so try and find the deck this goes with.
						string strDeckName = deck.FileName.ToUpper().Replace("_LAND_POOL", "");
						foreach (Deck dkTest in m_lstDecks)
						{
							if (dkTest.FileName.Equals(strDeckName, StringComparison.OrdinalIgnoreCase))
							{
								dkTest.LandPool = deck;
								break;
							}
						}
					}
				}

				// Now that all the decks have been loaded now we need to load and associate their unlocks.
				//	Additionally update the used ids.
				foreach (WadBase wad in m_hsetWads)
				{
					Dictionary<int, string> dicIds = wad.LoadUnlocks(this);
					foreach (KeyValuePair<int, string> kvId in dicIds)
					{
						if (m_dicUsedIds.ContainsKey(kvId.Key))
						{
							// Only report if the filenames aren't the same, if they are the same it is due to one overriding another.
							if (!kvId.Value.Equals(m_dicUsedIds[kvId.Key]))
								Settings.ReportError(null, ErrorPriority.Low, kvId.Value + " uses a UID that conflicts with " + m_dicUsedIds[kvId.Key] + ".");
						}
						else
							m_dicUsedIds.Add(kvId.Key, kvId.Value);
					}
				}
			}
		}
	}
}
